use io;

/* Read instance data */
function input() {
    COLOR_HIGH_LOW = 0;
    HIGH_LOW_COLOR = 1;
    HIGH_COLOR_LOW = 2;
    COLOR_HIGH = 3;
    HIGH_COLOR = 4;
    
    local usage = "Usage: localsolver car_sequencing_color.lsp "
            + "inFileName=inputFile [solFileName=outputFile] [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;
    local inFile = io.openRead(inFileName);
    nbPositions = inFile.readInt();
    nbOptions = inFile.readInt();
    nbClasses = inFile.readInt();
    paintBatchLimit = inFile.readInt();
    objectiveOrder = inFile.readInt();
    startPosition = inFile.readInt();

    hasLowPriorityOptions = false;
    for [o in 0...nbOptions] {
        maxCarsPerWindow[o] = inFile.readInt();
        windowSize[o] = inFile.readInt();
        isPriorityOption[o] = (inFile.readInt() == 1);
        if (!isPriorityOption[o]) hasLowPriorityOptions = true;
    }

    if (!hasLowPriorityOptions) {
        if (objectiveOrder == COLOR_HIGH_LOW) objectiveOrder = COLOR_HIGH;
        else if (objectiveOrder == HIGH_COLOR_LOW) objectiveOrder = HIGH_COLOR;
        else if (objectiveOrder == HIGH_LOW_COLOR) objectiveOrder = HIGH_COLOR;
    }
    
    for [c in 0...nbClasses] {
        colorClass[c] = inFile.readInt();
        nbCars[c] = inFile.readInt();
        options[c][o in 0...nbOptions] = (inFile.readInt()==1);
    }
    
    initialSequence[p in 0...nbPositions] = inFile.readInt();
}

/* Declare the optimization model */
function model() {
    // sequence[i] = j if class initially planned on position j is produced on position i
    sequence <- list(nbPositions);
    
    // sequence is a permutation of the initial production plan, all indexes must
    // appear exactly once
    constraint partition(sequence);

    // Past classes (before startPosition) can not move
    for [p in 0...startPosition]
        constraint sequence[p] == p;

    // Number of cars with option o in each window
    nbCarsWindows[o in 0...nbOptions][j in startPosition-windowSize[o]+1...nbPositions] 
            <- sum[k in 0...windowSize[o] : j + k >= 0 && j + k < nbPositions]
            (options[initialSequence[sequence[j + k]]][o]);
    
    // Number of violations of option o capacity in each window
    nbViolationsWindows[o in 0...nbOptions] 
            <- sum[j in startPosition-windowSize[o]+1...nbPositions]
            (max(0, nbCarsWindows[o][j] - maxCarsPerWindow[o]));

    objectiveHighPriority <- sum[o in 0...nbOptions : isPriorityOption[o]](nbViolationsWindows[o]);
    objectiveLowPriority <- sum[o in 0...nbOptions : !isPriorityOption[o]](nbViolationsWindows[o]);

    // Color change between position p and position p + 1
    colorChange[p in startPosition-1...nbPositions-1] 
            <- colorClass[initialSequence[sequence[p]]] != colorClass[initialSequence[sequence[p + 1]]];
    objectiveColor <- sum[p in startPosition-1...nbPositions-1](colorChange[p]);

    // Paint limit constraints: at least one change every paintBatchLimit positions
    for [p in startPosition...nbPositions-paintBatchLimit-1]
        constraint or[p2 in 0...paintBatchLimit](colorChange[p + p2]);

    // Declare the objectives in the correct order
    if(objectiveOrder == COLOR_HIGH_LOW) {
        obj[0] <- objectiveColor;
        obj[1] <- objectiveHighPriority;
        obj[2] <- objectiveLowPriority;
    } else if(objectiveOrder == HIGH_COLOR_LOW) {
        obj[0] <- objectiveHighPriority;
        obj[1] <- objectiveColor;
        obj[2] <- objectiveLowPriority;
    } else if(objectiveOrder == HIGH_LOW_COLOR) {
        obj[0] <- objectiveHighPriority;
        obj[1] <- objectiveLowPriority;
        obj[2] <- objectiveColor;       
    } else if(objectiveOrder == COLOR_HIGH) {
        obj[0] <- objectiveColor;
        obj[1] <- objectiveHighPriority;
    } else if(objectiveOrder == HIGH_COLOR) {
        obj[0] <- objectiveHighPriority;
        obj[1] <- objectiveColor;
    }
    
    for [o in obj] minimize o;
}

/* Parametrize the solver */
function param() {
    // Set the initial solution
    sequence.value.clear();
    for [p in 0...nbPositions]
        sequence.value.add(p);
    if (lsTimeLimit == nil) lsTimeLimit = 60;
}

/* Write the solution in a file with the following format: 
 * - 1st line: value of the objectives;
 * - 2nd line: for each position p, index of class at positions p. */
function output() {
    if (solFileName == nil) return;
    local solFile = io.openWrite(solFileName);
    solFile.print(objectiveColor.value, " ");
    solFile.print(objectiveHighPriority.value, " ");
    solFile.println(objectiveLowPriority.value);
    local listSolution = sequence.value;
    for [p in 0...nbPositions]
        solFile.print(initialSequence[listSolution[p]], " ");
    solFile.println();
}

