import localsolver
import sys

COLOR_HIGH_LOW = 0
HIGH_LOW_COLOR = 1
HIGH_COLOR_LOW = 2
COLOR_HIGH = 3
HIGH_COLOR = 4

def read_integers(filename):
    with open(filename) as f:
        return [int(elem) for elem in f.read().split()]

#
# Read instance data
#
def read_instance(instance_file):
    file_it = iter(read_integers(instance_file))
    nb_positions = next(file_it)
    nb_options = next(file_it)
    nb_classes = next(file_it)
    paint_batch_limit = next(file_it)
    objective_order = next(file_it)
    start_position = next(file_it)

    max_cars_per_window = []
    window_size = []
    is_priority_option = []
    has_low_priority_options = False

    for o in range(nb_options):
        max_cars_per_window.append(next(file_it))
        window_size.append(next(file_it))
        is_prio = next(file_it) == 1
        is_priority_option.append(is_prio)
        if not is_prio:
            has_low_priority_options = True

    if not has_low_priority_options:
        if objective_order == COLOR_HIGH_LOW:
            objective_order = COLOR_HIGH
        elif objective_order == HIGH_COLOR_LOW:
            objective_order = HIGH_COLOR
        elif objective_order == HIGH_LOW_COLOR:
            objective_order = HIGH_COLOR

    color_class = []
    nb_cars = []
    options_data = []

    for c in range(nb_classes):
        color_class.append(next(file_it))
        nb_cars.append(next(file_it))
        options_data.append([next(file_it) == 1 for i in range(nb_options)])

    initial_sequence = [next(file_it) for p in range(nb_positions)]

    return nb_positions, nb_options, paint_batch_limit, objective_order, start_position, \
        max_cars_per_window, window_size, is_priority_option, has_low_priority_options, \
        color_class, options_data, initial_sequence

def main(instance_file, output_file, time_limit):
    nb_positions, nb_options, paint_batch_limit, objective_order, start_position, \
        max_cars_per_window, window_size, is_priority_option, has_low_priority_options, \
        color_class, options_data, initial_sequence = read_instance(instance_file)

    with localsolver.LocalSolver() as ls:
        #
        # Declare the optimization model
        #
        model = ls.model

        # sequence[i] = j if class initially planned on position j is produced on position i
        sequence = model.list(nb_positions)

        # sequence is a permutation of the initial production plan, all indexes must appear
        # exactly once
        model.constraint(model.partition(sequence))

        # Past classes (before startPosition) can not move
        [model.constraint(sequence[p] == p) for p in range(start_position)]

        # Create LocalSolver arrays to be able to access them with "at" operators
        initials = model.array(initial_sequence)
        colors = model.array(color_class)
        options = model.array(options_data)

        # Number of cars with option o in each window
        nb_cars_windows = [None] * nb_options
        for o in range(nb_options):
            nb_cars_windows[o] = [None] * nb_positions
            for j in range(start_position - window_size[o] + 1, nb_positions):
                nb_cars_windows[o][j] = model.sum()
                for k in range(window_size[o]):
                    if j + k >= 0 and j + k < nb_positions:
                        class_at_position = initials[sequence[j + k]]
                        nb_cars_windows[o][j].add_operand(model.at(
                            options,
                            class_at_position,
                            o))

        # Number of violations of option o capacity in each window
        objective_high_priority = model.sum()
        if has_low_priority_options:
            objective_low_priority = model.sum()

        for o in range(nb_options):
            nb_violations_windows = model.sum(
                model.max(
                    nb_cars_windows[o][p] - max_cars_per_window[o], 0)
                    for p in range(start_position - window_size[o] + 1, nb_positions))
            if is_priority_option[o]:
                objective_high_priority.add_operand(nb_violations_windows)
            else:
                objective_low_priority.add_operand(nb_violations_windows)

        # Color change between position p and position p + 1
        color_change = [None] * (nb_positions - 1)
        objective_color = model.sum()
        for p in range(start_position - 1, nb_positions - 1):
            current_class = initials[sequence[p]]
            next_class = initials[sequence[p + 1]]
            color_change[p] = colors[current_class] != colors[next_class]
            objective_color.add_operand(color_change[p])

        # Paint limit constraints: at least one change every paintBatchLimit positions
        for p in range(start_position, nb_positions - paint_batch_limit - 1):
            node_or = model.or_(color_change[p + p2] for p2 in range(paint_batch_limit))
            model.constraint(node_or)

        # Declare the objectives in the correct order
        if objective_order == COLOR_HIGH_LOW:
            model.minimize(objective_color)
            model.minimize(objective_high_priority)
            model.minimize(objective_low_priority)
        elif objective_order == HIGH_COLOR_LOW:
            model.minimize(objective_high_priority)
            model.minimize(objective_color)
            model.minimize(objective_low_priority)
        elif objective_order == HIGH_LOW_COLOR:
            model.minimize(objective_high_priority)
            model.minimize(objective_low_priority)
            model.minimize(objective_color)
        elif objective_order == COLOR_HIGH:
            model.minimize(objective_color)
            model.minimize(objective_high_priority)
        elif objective_order == HIGH_COLOR:
            model.minimize(objective_high_priority)
            model.minimize(objective_color)

        model.close()


        # Set the initial solution
        sequence.get_value().clear()
        for p in range(nb_positions):
            sequence.get_value().add(p)

        # Parameterize the solver
        ls.param.time_limit = time_limit

        ls.solve()

        #
        # Write the solution in a file with the following format:
        # - 1st line: value of the objectives;
        # - 2nd line: for each position p, index of class at positions p.
        #
        if output_file is not None:
            with open(output_file, 'w') as f:
                f.write("%d " % objective_color.value)
                f.write("%d " % objective_high_priority.value)
                f.write("%d\n" % objective_low_priority.value)
                for p in range(nb_positions):
                    f.write("%d " % sequence.value[p])

                f.write("\n")

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("Usage: python car_sequencing_color.py instance_file [output_file] [time_limit]")
        sys.exit(1)

    instance_file = sys.argv[1]
    output_file = sys.argv[2] if len(sys.argv) >= 3 else None
    time_limit = int(sys.argv[3]) if len(sys.argv) >= 4 else 60
    main(instance_file, output_file, time_limit)